from promptim.trainer import PromptWrapper, Task


def exact_match(run, example):
    """Evaluate the exact match correctness of the NLI result."""
    try:
        predicted = run.outputs["is_entailed"]
        expected = example.outputs["answer"]
        score = expected.lower() == predicted.lower()
    except Exception:
        try:
            expected = example.outputs["answer"]
            expected_bool = {"no": False, "yes": True}.get(expected.strip().lower())
            score = run.outputs["output"].is_entailed == expected_bool
        except Exception:
            score = 0
    return {
        "key": "exact_match",
        "score": int(score),
    }


scone_task = Task(
    name="Scone (NLI)",
    dataset="scone-optim",
    initial_prompt=PromptWrapper(identifier="langchain-ai/scone-example:d49910d6"),
    evaluators=[exact_match],
    evaluator_descriptions={
        "exact_match": "Directly compares the expected against the predicted outputs. 1 if correct, 0 if incorrect."
    },
)



import functools
import logging

from langchain_openai import ChatOpenAI
from promptim.trainer import PromptWrapper, Task
from pydantic import BaseModel, Field


logger = logging.getLogger(__name__)


class Grade(BaseModel):
    """Call to submit your grade."""

    reasoning: str = Field(
        description="First, explain your thought process on why you are giving the provided grade."
    )
    score: int = Field(
        ge=0, le=5, description="Then, submit your score on a scale from 0 to 5."
    )

    @property
    def normalized(self):
        return self.score / 5


@functools.lru_cache
def _get_judge():
    from trustcall import create_extractor

    return create_extractor(
        ChatOpenAI(model="gpt-4o-mini"), tools=[Grade], tool_choice=Grade.__name__
    )


utemplate = """Grade the following:
Predicted: {predicted}
Reference example: {reference}"""


async def summary_quality(run, example):
    predicted = run.outputs.get("summary")
    rubric = """Grade the quality of summary. If it fails any criteria, give a 0. If it's perfect, give a 5.
Criteria:
- Must not include idle words like "the email is about X"
- Preferred format is <person> from <org> needs/wants X
"""
    reference = example.outputs["summary"]
    result = await _get_judge().ainvoke(
        [
            ("system", rubric),
            ("user", utemplate.format(predicted=predicted, reference=reference)),
        ]
    )
    grade: Grade = result["responses"][0]
    pf = "Pass" if grade.score >= 4 else "Fail"
    return {"score": grade.normalized, "comment": f"{pf}: {grade.reasoning}"}


def accuracy_check(run, example, key: str):
    predicted = run.outputs.get(key)
    reference = example.outputs.get(key)
    if reference is None:
        return {
            "key": f"{key}-correctness",
            "comment": "Skipping - reference label not found.",
        }
    score = (
        predicted == reference
        if not isinstance(reference, list)
        else predicted in reference
    )
    pf = "Pass" if score else "Fail"
    return {
        "key": f"{key}-correctness",
        "score": score,
        "comment": f"{pf}",
    }  #: Expected {reference}. Got: {predicted}. Why did you get this wrong? Think deeply and update associations."}


classifiers = [
    functools.partial(accuracy_check, key=key)
    for key in [
        "category",
        "support_category",
        "ticket_status",
        "requires_response",
        "non_support_category",
    ]
]
evaluators = [summary_quality, *classifiers]


ticket_classification_task = Task(
    name="Ticket Classification",
    description="A task to classify customer support tickets",
    dataset="ticket-classification-optim",
    initial_prompt=PromptWrapper(
        identifier="langchain-ai/ticket-classifier-example:376ab5e4",
        which=1,
    ),
    evaluators=evaluators,
    evaluator_descriptions={
        "summary_quality": "Evaluates the quality of the summary",
        "category-correctness": "Checks if the category is correct",
        "support_category-correctness": "Checks if the support category is correct",
        "ticket_status-correctness": "Checks if the ticket status is correct",
        "requires_response-correctness": "Checks if the requires_response field is correct",
        "non_support_category-correctness": "Checks if the non-support category is correct",
    },
)

if __name__ == "__main__":
    import random

    import langsmith as ls

    c = ls.Client()
    examples = list(
        c.list_examples(dataset_name="customer-support-bot.test_extraction")
    )

    random.shuffle(examples)
    full = examples.copy()
    train, dev, test = [], [], []
    dname = "ticket-classification-optim"
    try:
        dataset = c.create_dataset(dataset_name=dname)
    except Exception:
        c.delete_dataset(dataset_name=dname)
        dataset = c.create_dataset(dataset_name=dname)
    for ds, size, name in zip(
        [train, dev, test], [41, 20, 20], ["train", "dev", "test"]
    ):
        for i in range(size):
            ds.append(full.pop())

        outputs = [e.outputs for e in ds]
        for o in outputs:
            for k, v in o.pop("outputs", {}).items():
                o[k] = v
        c.create_examples(
            inputs=[e.inputs for e in ds],
            outputs=outputs,
            dataset_id=dataset.id,
            splits=[name] * len(ds),
        )



import langsmith as ls
from langchain_anthropic import ChatAnthropic
from promptim.trainer import (
    PromptWrapper,
    PromptOptimizer,
    Task,
    DEFAULT_METAPROMPT,
)
from promptim.tasks.scone import scone_task
from promptim.tasks.tweet_generator import tweet_task
from promptim.tasks.simpleqa import simpleqa_task
from promptim.tasks.ticket_classification import ticket_classification_task
from langchain_core.prompts import ChatPromptTemplate


DEFAULT_METAMETAPROMPT = """You are an expert in prompt optimization systems. Your task is to improve the effectiveness of prompt optimization prompts - the prompts used to guide the improvement of task-specific prompts.

Current Optimization Prompt:
{current_prompt}

Performance Data:
Shows how this optimization prompt performed in improving various task-specific prompts
{annotated_results}

Objective:
Improve the optimization prompt to better guide the enhancement of task-specific prompts across:
- Systematic analysis of prompt performance
- Identification of improvement opportunities
- Generation of enhanced prompts
- Validation of improvements

Analysis Steps:
1. Optimization Effectiveness
   - How well did this optimization prompt guide improvements?
   - Which aspects of prompt optimization were handled well/poorly?
   - What patterns emerge in successful vs unsuccessful optimization attempts?

2. Structural Assessment
   - How clearly does it guide the optimization process?
   - How well does it maintain prompt constraints?
   - What components are missing or ineffective?

3. Improvement Strategy
   - Which elements of the optimization process need enhancement?
   - How can we make the optimization guidance more effective?
   - What additional checks or validations would help?

Output Format:
<effectiveness_analysis>
Analysis of how well this optimization prompt guides improvements
</effectiveness_analysis>

<improvement_strategy>
Specific changes to enhance optimization capabilities
</improvement_strategy>

<improved_optimization_prompt>
The enhanced prompt for optimizing task-specific prompts
</improved_optimization_prompt>"""


class MetapromptSystem:
    """System for running the metaprompt optimization task."""

    def __init__(self, task_map: dict[str, Task], meta_prompt: PromptWrapper):
        self.task_map = task_map
        self.model = ChatAnthropic(
            model="claude-3-5-sonnet-20241022", max_tokens_to_sample=8192
        )
        self.trainer = PromptOptimizer(self.model, meta_prompt.get_prompt_str())

    async def __call__(self, prompt: ChatPromptTemplate, inputs: dict) -> dict:
        task = self.task_map[inputs["task"]]
        task.initial_prompt.load()

        # Run initial prompt on batch using aevaluate
        async def predict(example_inputs: dict):
            return await task.system_safe(task.initial_prompt.load(), example_inputs)

        train_batch = list(
            self.trainer.client.list_examples(example_ids=inputs["train_batch"])
        )
        dev_batch = list(
            self.trainer.client.list_examples(example_ids=inputs["dev_batch"])
        )
        initial_results = [
            r
            async for r in (
                await ls.aevaluate(
                    predict,
                    data=train_batch,
                    evaluators=task.evaluators,
                )
            )
        ]
        task.initial_prompt.get_prompt_str()

        # Generate new downstream task prompt
        extracted = await self.trainer.apply_metaprompt(
            current_prompt=task.initial_prompt,
            meta_prompt=prompt.messages[0].prompt.template,  # type: ignore
            task=task,
            results=initial_results,
        )

        # Now we actually evaluate based on how well the updated prompt's "improvements"
        # translate to a dev batch
        initial_dev_results = [
            r
            async for r in (
                await ls.aevaluate(
                    predict,
                    data=dev_batch,
                    evaluators=task.evaluators,
                )
            )
        ]
        initial_dev_scores = await self.trainer.calculate_scores(initial_dev_results)

        async def predict_new(example_inputs: dict):
            return await task.system_safe(extracted._cached, example_inputs)

        new_results = [
            r
            async for r in (
                await ls.aevaluate(
                    predict_new,
                    data=dev_batch,
                    evaluators=task.evaluators,
                )
            )
        ]
        new_scores = await self.trainer.calculate_scores(new_results)
        return {
            "original_prompt": task.initial_prompt,
            "new_prompt": extracted.get_prompt_str(),
            # "reasoning_for_changes": extracted.analysis,
            "initial_scores": initial_dev_scores,
            "new_scores": new_scores,
        }


def metaprompt_evaluator(run, example):
    """Evaluate the performance of the metaprompt."""
    original_score = sum(run.outputs["initial_scores"].values()) / len(
        run.outputs["initial_scores"]
    )
    new_score = sum(run.outputs["new_scores"].values()) / len(run.outputs["new_scores"])
    # Map the difference in scores to a 0 to 1 scale
    score_diff = new_score - original_score
    normalized_score = max(0, min(1, (score_diff + 1) / 2))
    if normalized_score > 0.5:
        comment = "The average scores improved after making the changes suggested by the prompt optimizer."
    elif normalized_score < 0.5:
        comment = "The average score dropped after making the changes suggested by the prompt optimizer."
    else:
        comment = "The average score remained the same after making the changes suggested by the prompt optimizer."

    return {
        "key": "metaprompt_improvement",
        "score": normalized_score,
        "comment": comment,
    }


prompt_config = PromptWrapper(prompt_str=DEFAULT_METAPROMPT)
metaprompt_task = Task(
    name="Metaprompt Optimizer",
    description="A meta-optimization task that aims to improve the prompt used for optimizing task-specific prompts. This task evaluates and enhances the effectiveness of the prompt optimization process itself, leading to better performance across various language tasks.",
    dataset="metaprompt-optim",
    initial_prompt=prompt_config,
    evaluators=[metaprompt_evaluator],
    evaluator_descriptions={
        "metaprompt_improvement": "Checks if the new prompt leads to improved scores. 1 if better, 0.5 if same, 0 if worse."
    },
    system=MetapromptSystem(
        {
            "scone": scone_task,
            "tweet": tweet_task,
            "simpleqa": simpleqa_task,
            "ticket_classification_task": ticket_classification_task,
        },
        meta_prompt=prompt_config,
    ),
)


if __name__ == "__main__":
    import argparse
    from langsmith import Client
    import random

    random.seed(42)

    def create_datasets(client, tasks, batchsize, overwrite=False):
        datasets = {}
        for split_name in ["train", "dev", "test"]:
            dataset_name = "metaprompt-optim"
            if overwrite:
                if client.has_dataset(dataset_name=dataset_name):
                    client.delete_dataset(dataset_name=dataset_name)
            datasets[split_name] = client.create_dataset(dataset_name=dataset_name)

        for task_name in tasks:
            task_datasets = {
                "train": client.list_examples(dataset_name=f"{task_name}-train"),
                "dev": client.list_examples(dataset_name=f"{task_name}-dev"),
                "test": client.list_examples(dataset_name=f"{task_name}-test"),
            }

            for split_name in ["train", "dev", "test"]:
                examples = []
                task_split_examples = list(task_datasets[split_name])
                random.shuffle(task_split_examples)
                for i in range(len(task_split_examples) // (batchsize * 2)):
                    batch = [
                        str(example.id)
                        for example in task_split_examples[i : i + batchsize * 2]
                    ]

                    examples.append(
                        {
                            "task": task_name,
                            "train_batch": batch[0 : len(batch) // 2],
                            "dev_batch": batch[len(batch) // 2 :],
                        }
                    )

                client.create_examples(
                    inputs=examples,
                    dataset_id=datasets[split_name].id,
                    splits=[split_name] * len(examples),
                )

        return datasets

    parser = argparse.ArgumentParser(
        description="Generate datasets for metaprompt optimization"
    )
    parser.add_argument(
        "--batchsize", type=int, default=5, help="Number of examples in each batch."
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing datasets if they exist.",
    )
    args = parser.parse_args()

    client = Client()
    tasks = ["scone", "tweet", "simpleqa", "ticket_classification_task"]

    datasets = create_datasets(client, tasks, args.batchsize, args.overwrite)

    print("Datasets created successfully!")
    for name, dataset in datasets.items():
        print(f"{name.capitalize()} dataset ID: {dataset.id}")



import asyncio
import importlib.util
import json
import os
from typing import Optional, cast

import click
import langsmith as ls
from langsmith.utils import LangSmithNotFoundError
import sys


def get_tasks(task_name: str):
    from promptim.tasks.metaprompt import metaprompt_task
    from promptim.tasks.scone import scone_task
    from promptim.tasks.simpleqa import simpleqa_task
    from promptim.tasks.ticket_classification import ticket_classification_task
    from promptim.tasks.tweet_generator import tweet_task

    tasks = {
        "scone": scone_task,
        "tweet": tweet_task,
        "metaprompt": metaprompt_task,
        "simpleqa": simpleqa_task,
        "ticket-classification": ticket_classification_task,
    }
    return tasks.get(task_name)


def load_task(name_or_path: str):
    from promptim.trainer import Task

    task = get_tasks(name_or_path)
    if task:
        return task, {}
    # If task is not in predefined tasks, try to load from file
    try:
        with open(name_or_path, "r", encoding="utf-8") as f:
            config = json.load(f)
        if "$schema" in config:
            del config["$schema"]
        evaluators_path = config["evaluators"]
        module_path, evaluators_variable = [
            part for part in evaluators_path.split(":") if part
        ]
        # First try to load it relative to the config path
        config_dir = os.path.dirname(name_or_path)
        relative_module_path = os.path.join(config_dir, module_path)
        if os.path.exists(relative_module_path):
            module_path = relative_module_path
        spec = importlib.util.spec_from_file_location("evaluators_module", module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        evaluators = getattr(module, evaluators_variable)
        if not isinstance(evaluators, list):
            raise ValueError(
                f"Expected evaluators to be a list, but got {type(evaluators).__name__}"
            )
        task = Task.from_dict({**config, "evaluators": evaluators})
        return task, config
    except Exception as e:
        raise ValueError(f"Could not load task from {name_or_path}: {e}")


async def run(
    task_name: str,
    batch_size: int,
    train_size: int,
    epochs: int,
    annotation_queue: Optional[str] = None,
    debug: bool = False,
    commit: bool = True,
):
    task, config = load_task(task_name)
    from promptim.trainer import PromptOptimizer

    optimizer = PromptOptimizer.from_config(config.get("optimizer_config", {}))

    with ls.tracing_context(project_name="Optim"):
        prompt, score = await optimizer.optimize_prompt(
            task,
            batch_size=batch_size,
            train_size=train_size,
            epochs=epochs,
            annotation_queue=annotation_queue,
            debug=debug,
            commit_prompts=commit,
        )
    if commit and task.initial_prompt.identifier is not None:
        prompt.push_prompt(
            identifier=task.initial_prompt.identifier.rsplit(":", maxsplit=1)[0],
            include_model_info=True,
            client=optimizer.client,
        )

    return prompt, score


@click.group()
@click.version_option(version="1")
def cli():
    """Optimize prompts for different tasks."""
    pass


@cli.command()
@click.option(
    "--task",
    help="Task to optimize. You can pick one off the shelf or select a path to a config file. "
    "Example: 'examples/tweet_writer/config.json",
)
@click.option("--batch-size", type=int, default=40, help="Batch size for optimization")
@click.option(
    "--train-size", type=int, default=40, help="Training size for optimization"
)
@click.option("--epochs", type=int, default=2, help="Number of epochs for optimization")
@click.option("--debug", is_flag=True, help="Enable debug mode")
@click.option(
    "--annotation-queue",
    type=str,
    default=None,
    help="The name of the annotation queue to use. Note: we will delete the queue whenever you resume training (on every batch).",
)
@click.option(
    "--no-commit",
    is_flag=True,
    help="Do not commit the optimized prompt to the hub",
)
def train(
    task: str,
    batch_size: int,
    train_size: int,
    epochs: int,
    debug: bool,
    annotation_queue: Optional[str],
    no_commit: bool,
):
    """Train and optimize prompts for different tasks."""
    results = asyncio.run(
        run(
            task,
            batch_size,
            train_size,
            epochs,
            annotation_queue,
            debug,
            commit=not no_commit,
        )
    )
    prompt_config, score = results
    print("Final\n\n")
    print(prompt_config.get_prompt_str())
    print("\n\n")
    print(f"Identifier: {prompt_config.identifier}")
    print(f"Score: {score}")


@cli.group()
def create():
    """Commands for creating new tasks and examples."""
    pass


class MissingPromptError(ValueError):
    """Error raised when a prompt is not found."""

    def __init__(self, attempted: str):
        self.attempted = attempted
        super().__init__(f"Prompt not found: {attempted}")


def _try_get_prompt(client, prompt: str | None):
    from langsmith import Client
    from langchain_core.prompts import ChatPromptTemplate
    from langchain_core.runnables import RunnableSequence, RunnableBinding
    from promptim.trainer import PromptWrapper

    expected_run_outputs = 'predicted: AIMessage = run.outputs["output"]'
    client = cast(Client, client)
    if prompt is None:
        prompt = click.prompt(
            "Enter the identifier for the initial prompt\n"
            "\tFormat: prompt-name"
            " OR <organization>/<prompt-name>:<commit-or-tag>\n"
            "\tExamples:\n"
            "\t  - langchain-ai/tweet-generator-example-with-nothing:starter\n"
            "\t  - langchain-ai/tweet-generator-example:c39837bd\n"
            "Prompt identifier"
        )
        if prompt == "q":
            click.echo("Exiting task creation.")
            sys.exit()

    # Fetch prompt
    try:
        prompt_repo = client.get_prompt(prompt)
        chain = client.pull_prompt(prompt, include_model=True)
    except LangSmithNotFoundError:
        raise MissingPromptError(attempted=prompt)

    if isinstance(chain, ChatPromptTemplate):
        prompt_obj = chain
    elif isinstance(chain, RunnableSequence):
        prompt_obj = chain.first
    else:
        raise ValueError(f"Unrecognized prompt format: {chain}")
    if isinstance(prompt_obj, ChatPromptTemplate):
        pass
    elif (
        isinstance(chain, RunnableSequence)
        and isinstance(chain.steps[1], RunnableBinding)
        and chain.steps[1].kwargs.get("tools")
    ):
        tools = chain.steps[1].kwargs.get("tools")
        tool_names = [
            t.get("function", {}).get("name")
            for t in tools
            if t.get("function", {}).get("name")
        ]
        expected_run_outputs = f"# AI message contains optional tool_calls from your prompt\n    # Example tool names: {tool_names}\n    {expected_run_outputs}"
    else:
        raise ValueError(f"Unexpected prompt type: {type(prompt_obj)}\n\n{prompt_obj}")
    identifier = prompt
    if "/" in identifier:  # It may be a public prompt:
        # CRAZY HANDLING I HATE THIS
        tenant_id = client._get_tenant_id()
        if prompt_repo.tenant_id != str(tenant_id):
            # Warn user and ask for confirmation to clone the prompt
            click.echo(
                f"Warning: The prompt '{identifier}' does not belong to your workspace."
            )
            truncated_identifier = identifier.split("/", maxsplit=1)[1]
            target_repo = truncated_identifier.split(":")[0]

            # Check if target repo exists
            try:
                client.pull_prompt_commit(target_repo)
                repo_exists = True
            except LangSmithNotFoundError:
                repo_exists = False

            if repo_exists:
                # Check if truncated_identifier exists
                try:
                    client.pull_prompt_commit(truncated_identifier)
                    click.echo(f"Using existing prompt: {truncated_identifier}")
                    identifier = truncated_identifier
                except LangSmithNotFoundError:
                    click.echo(
                        f"Prompt {truncated_identifier} not found. Using {target_repo} instead."
                    )
                    identifier = target_repo
            else:
                clone_confirmation = click.confirm(
                    f"Would you like to clone prompt {target_repo} to your workspace before continuing?",
                    default=True,
                )

                if clone_confirmation:
                    try:
                        if isinstance(chain, RunnableSequence):
                            cloned_prompt = PromptWrapper._push_seq(
                                client, chain, identifier=truncated_identifier
                            )
                        else:
                            cloned_prompt = client.push_prompt(
                                truncated_identifier, object=prompt_obj
                            )
                        identifier = cloned_prompt.split("?")[0].split(
                            "/prompts/", maxsplit=1
                        )[1]
                        identifier = ":".join(identifier.rsplit("/"))
                        click.echo(
                            f"Prompt cloned successfully to {cloned_prompt}. New identifier: {identifier}"
                        )
                    except Exception as e:
                        click.echo(f"Error cloning prompt: {e}")
                        click.echo(f"Continuing with the original prompt {identifier}.")
                        click.echo(
                            "You will have to clone this manually in the UI if you want to push optimized commits."
                        )
                else:
                    click.echo(f"Continuing with the original prompt {identifier}.")
                    click.echo(
                        "You will have to clone this manually in the UI if you want to push optimized commits."
                    )

    return prompt_obj, identifier, expected_run_outputs


def get_prompt(client, prompt: str | None):
    from langsmith import Client

    client = cast(Client, client)
    while True:
        try:
            return _try_get_prompt(client, prompt)
        except MissingPromptError as e:
            click.echo(f"Could not find prompt: {e.attempted}")
            response = client.list_prompts(
                query=e.attempted.split(":")[0].strip(),
                match_prefix=True,
                limit=10,
                has_commits=True,
            )
            matches = []
            for repo in response.repos:
                if repo.last_commit_hash:
                    matches.append(f"{repo.repo_handle}:{repo.last_commit_hash[:8]}")
            if not matches:
                prompt = None
                click.echo("Please try again or press 'q' to quit.")
            else:
                click.echo("Did you mean one of these?")
                for i, match in enumerate(matches, 1):
                    click.echo(f"{i}. {match}")
                selection = click.prompt(
                    "Enter the number of your selection or type an identifier to try again"
                )
                if selection.isdigit() and 1 <= int(selection) <= len(matches):
                    prompt = matches[int(selection) - 1]
                elif selection.strip() == "q":
                    sys.exit()
                else:
                    prompt = selection.strip() or None
                    click.echo("Please try again or press 'q' to quit.")
        except click.Abort:
            sys.exit()
        except Exception as e:
            click.echo(f"Error loading prompt: {e!r}")
            click.echo("Please try again or press 'q' to quit.")
            prompt = None


class MissingDatasetError(ValueError):
    """Error raised when a dataset is not found."""

    def __init__(self, attempted: str):
        self.attempted = attempted
        super().__init__(f"Dataset not found: {attempted}")


def get_dataset(client, dataset: str | None):
    from langsmith import Client

    client = cast(Client, client)
    while True:
        try:
            return _try_get_dataset(client, dataset)
        except MissingDatasetError as e:
            print(f"Could not find dataset: {e.attempted}")
            response = client.list_datasets(
                dataset_name_contains=e.attempted,
                limit=10,
            )
            matches = [ds.name for ds in response]
            if not matches:
                create_dataset = click.confirm(
                    f"Dataset '{e.attempted}' not found. Would you like to create it?",
                    default=False,
                )
                if create_dataset:
                    ds = client.create_dataset(dataset_name=e.attempted)
                    click.echo(f"Dataset '{e.attempted}' created successfully.")
                    return ds
                else:
                    dataset = None
                    print("Please try again or press 'q' to quit.")
            else:
                print("Did you mean one of these?")
                for i, match in enumerate(matches, 1):
                    print(f"{i}. {match}")
                print(f"{len(matches) + 1}. Create a new dataset")
                selection = click.prompt(
                    "Enter the number of your selection, type a name to try again, or choose to create a new dataset"
                )
                if selection.isdigit():
                    if 1 <= int(selection) <= len(matches):
                        dataset = matches[int(selection) - 1]
                    elif int(selection) == len(matches) + 1:
                        new_name = click.prompt("Enter the name for the new dataset")
                        ds = client.create_dataset(dataset_name=new_name)
                        click.echo(f"Dataset '{new_name}' created successfully.")
                        return ds
                elif selection.strip() == "q":
                    sys.exit()
                else:
                    dataset = selection.strip() or None
                    print("Please try again or press 'q' to quit.")
        except click.Abort:
            sys.exit()
        except Exception as e:
            print(f"Error loading dataset: {e!r}")
            print("Please try again or press 'q' to quit.")
            dataset = None


def _try_get_dataset(client, dataset: str | None):
    from langsmith import Client

    client = cast(Client, client)
    if dataset is None:
        dataset = click.prompt(
            "Enter the name of an existing dataset or a URL of a public dataset:\n"
            "\tExamples:\n"
            "\t  - my-dataset\n"
            "\t  - https://smith.langchain.com/public/6ed521df-c0d8-42b7-a0db-48dd73a0c680/d\n"
            "Dataset name or URL"
        )
        if dataset == "q":
            print("Exiting task creation.")
            sys.exit()

    if dataset.startswith("https://"):
        ds = client.clone_public_dataset(dataset)
        return ds

    try:
        ds = client.read_dataset(dataset_name=dataset)
        return ds
    except LangSmithNotFoundError:
        raise MissingDatasetError(attempted=dataset)
    except Exception as e:
        raise ValueError(f"Could not fetch dataset '{dataset}': {e}") from e


@create.command("task")
@click.argument("path", type=click.Path(file_okay=False, dir_okay=True))
@click.option("--name", required=False, help="Name for the task.")
@click.option("--prompt", required=False, help="Name of the prompt in LangSmith")
@click.option(
    "--description", required=False, help="Description of the task for the optimizer."
)
@click.option("--dataset", required=False, help="Name of the dataset in LangSmith")
def create_task(
    path: str,
    name: str | None = None,
    prompt: str | None = None,
    dataset: str | None = None,
    description: str | None = None,
):
    """Create a new task directory with config.json and task file for a custom prompt and dataset."""
    from langsmith import Client

    client = Client()
    if not client.api_key:
        raise ValueError("LANGSMITH_API_KEY required to create the task.")
    if name is None:
        default_name = os.path.basename(os.path.abspath(path))
        name = click.prompt(
            "Enter a name for your task",
            default=default_name,
        ).strip()
        if name == "q":
            print("Exiting task creation.")
            return

    expected_imports = "from langchain_core.messages import AIMessage"
    prompt_obj, identifier, expected_run_outputs = get_prompt(client, prompt)

    # Create task directory
    os.makedirs(path, exist_ok=True)
    ds = get_dataset(client, dataset)
    try:
        example = next(client.list_examples(dataset_id=ds.id, limit=1))
    except Exception:
        example = None

    def json_comment(d: dict, indent: int = 4):
        return "\n".join(
            f"{' ' * indent}# {line}" for line in json.dumps(d, indent=2).splitlines()
        )

    example_content = ""
    if example is not None:
        example_inputs = json_comment(example.inputs) if example else None
        example_outputs = (
            (
                json_comment(example.outputs)
                if example.outputs is not None
                else "    # None"
            )
            if example
            else None
        )
        example_content = f"""
    # We've copied the inputs & outputs for the first example in the configured dataset.
    prompt_inputs = example.inputs
{example_inputs}
    reference_outputs = example.outputs # aka labels
{example_outputs}
    # The comments above autogenerated from example:
    # {example.url}
"""
    if description is None:
        description = click.prompt("Please enter a description for the task")

    # Create config.json
    config = {
        "name": name,
        "dataset": ds.name,
        "description": description,
        "evaluators": "./task.py:evaluators",
        "evaluator_descriptions": {
            "my_example_criterion": "CHANGEME: This is a description of what the example criterion is testing."
            " It is provided to the metaprompt "
            "to improve how it responds to different results."
        },
        "optimizer": {
            "model": {
                "model": "claude-3-5-sonnet-20241022",
                "max_tokens_to_sample": 8192,
            }
        },
        "initial_prompt": {
            "identifier": identifier,
            "model_config": {"model": "claude-3-5-haiku-20241022"},
        },
    }
    config[
        "$schema"
    ] = "https://raw.githubusercontent.com/hinthornw/promptimizer/refs/heads/main/config-schema.json"
    with open(os.path.join(path, "config.json"), "w") as f:
        json.dump(config, f, indent=2)

    # Create task.py with placeholder evaluators
    task_template = f"""\"\"\"Evaluators to optimize task: {name}.

THIS IS A TEMPLATE FOR YOU TO CHANGE!

Evaluators compute scores for prompts run over the configured dataset:
{ds.url}
\"\"\"
{expected_imports}
from langsmith.schemas import Run, Example

# Modify these evaluators to measure the requested criteria.
# For most prompt optimization tasks, run.outputs["output"] will contain an AIMessage
# (Advanced usage) If you are defining a custom system to optimize, then the outputs will contain the object returned by your system

def example_evaluator(run: Run, example: Example) -> dict:
    \"\"\"An example evaluator. Larger numbers are better.\"\"\"
    # The Example object contains the inputs and reference labels from a single row in your dataset (if provided).
    {example_content}    
    # The Run object contains the full trace of your system. Typically you run checks on the outputs,
    # often comparing them to the reference_outputs 
    {expected_run_outputs}

    # Implement your evaluation logic here
    score = len(str(predicted.content)) < 180  # Replace with actual score
    return {{
        # The evaluator keys here define the metric you are measuring
        # You can provide further descriptions for these in the config.json
        "key": "my_example_criterion",
        "score": score,
        "comment": (
            "CHANGEME: It's good practice to return "
            "information that can help the metaprompter fix mistakes, "
            "such as Pass/Fail or expected X but got Y, etc. "
            "This comment instructs the LLM how to improve."
            "The placeholder metric checks that the content is less than 180 in length."
        ),
    }}


evaluators = [example_evaluator]
"""

    with open(os.path.join(path, "task.py"), "w") as f:
        f.write(task_template.strip())

    print("*" * 80)
    print(f"Task '{name}' created at {path}")
    print(f"Config file created at: {os.path.join(path, 'config.json')}")
    print(f"Task file created at: {os.path.join(path, 'task.py')}")
    print(f"Using prompt:\n\n{prompt_obj.pretty_repr()}\n\n")
    print(f"Using dataset: {ds.url}")
    print(
        f"Remember to implement your custom evaluators in {os.path.join(path, 'task.py')}"
    )


@create.command("example")
@click.argument("path", type=click.Path(file_okay=False, dir_okay=True))
@click.argument("name", type=str)
def create_example_task(path: str, name: str):
    """Create an example task directory with config.json, task file, and example dataset."""
    # Create example dataset
    from langsmith import Client

    client = Client()
    if not client.api_key:
        raise ValueError("LANGSMITH_API_KEY required to create the example tweet task.")
    prompt = client.pull_prompt("langchain-ai/tweet-generator-example:c39837bd")
    identifier = f"{name}-starter"
    try:
        identifier = client.push_prompt(identifier, object=prompt, tags=["starter"])
    except ValueError as e:
        try:
            client.pull_prompt_commit(identifier)

        except Exception:
            raise e
        print(f"Prompt {name}-starter already found. Continuing.")

    identifier = identifier.split("?")[0].replace(
        "https://smith.langchain.com/prompts/", ""
    )
    identifier = identifier.rsplit("/", maxsplit=1)[0]
    identifier = f"{identifier}:starter"
    try:
        dataset = client.create_dataset(name)
    except Exception as e:
        if dataset := client.read_dataset(dataset_name=name):
            pass
        else:
            raise e

    topics = [
        "NBA",
        "NFL",
        "Movies",
        "Taylor Swift",
        "Artificial Intelligence",
        "Climate Change",
        "Space Exploration",
        "Cryptocurrency",
        "Healthy Living",
        "Travel Destinations",
        "Technology",
        "Fashion",
        "Music",
        "Politics",
        "Food",
        "Education",
        "Environment",
        "Science",
        "Business",
        "Health",
    ]

    for split_name, dataset_topics in [
        ("train", topics[:10]),
        ("dev", topics[10:15]),
        ("test", topics[15:]),
    ]:
        client.create_examples(
            inputs=[{"topic": topic} for topic in dataset_topics],
            dataset_id=dataset.id,
            splits=[split_name] * len(dataset_topics),
        )

    print(f"Task directory created at {path}")
    print(f"Example dataset '{dataset.name}' created with {len(topics)} examples")
    print(f"See: {dataset.url}")
    os.makedirs(path, exist_ok=True)

    config = {
        "name": "Tweet Generator",
        "dataset": name,
        "evaluators": "./task.py:evaluators",
        "optimizer": {
            "model": {
                "model": "claude-3-5-sonnet-20241022",
                "max_tokens_to_sample": 8192,
            }
        },
        "initial_prompt": {"identifier": identifier},
        "evaluator_descriptions": {
            "under_180_chars": "Checks if the tweet is under 180 characters. 1 if true, 0 if false.",
            "no_hashtags": "Checks if the tweet contains no hashtags. 1 if true, 0 if false.",
            "multiline": "Fails if the tweet is not multiple lines. 1 if true, 0 if false. 0 is bad.",
        },
    }

    config[
        "$schema"
    ] = "https://raw.githubusercontent.com/hinthornw/promptimizer/refs/heads/main/config-schema.json"
    with open(os.path.join(path, "config.json"), "w") as f:
        json.dump(config, f, indent=2)

    task_template = """
# You can replace these evaluators with your own.
# See https://docs.smith.langchain.com/evaluation/how_to_guides/evaluation/evaluate_llm_application#custom-evaluators
# for more information
def under_180_chars(run, example):
    \"\"\"Evaluate if the tweet is under 180 characters.\"\"\"
    result = run.outputs.get("tweet", "")
    score = int(len(result) < 180)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "under_180_chars",
        "score": score,
        "comment": comment,
    }

def no_hashtags(run, example):
    \"\"\"Evaluate if the tweet contains no hashtags.\"\"\"
    result = run.outputs.get("tweet", "")
    score = int("#" not in result)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "no_hashtags",
        "score": score,
        "comment": comment,
    }

evaluators = [multiple_lines, no_hashtags, under_180_chars]
"""

    with open(os.path.join(path, "task.py"), "w") as f:
        f.write(task_template.strip())


if __name__ == "__main__":
    cli()



from promptim.trainer import PromptWrapper, Task


def under_180_chars(run, example):
    """Evaluate if the tweet is under 180 characters."""
    result = run.outputs.get("tweet", "")
    score = int(len(result) < 180)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "under_180_chars",
        "score": score,
        "comment": comment,
    }


def no_hashtags(run, example):
    """Evaluate if the tweet contains no hashtags."""
    result = run.outputs.get("tweet", "")
    score = int("#" not in result)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "no_hashtags",
        "score": score,
        "comment": comment,
    }


def multiple_lines(run, example):
    """Evaluate if the tweet contains multiple lines."""
    result = run.outputs.get("tweet", "")
    score = int("\n" in result)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "multiline",
        "score": score,
        "comment": comment,
    }


tweet_task = Task(
    name="Tweet Generator",
    dataset="tweet-optim",
    initial_prompt=PromptWrapper(identifier="tweet-generator-example:c39837bd"),
    evaluators=[under_180_chars, no_hashtags, multiple_lines],
    evaluator_descriptions={
        "under_180_chars": "Checks if the tweet is under 180 characters. 1 if true, 0 if false.",
        "no_hashtags": "Checks if the tweet contains no hashtags. 1 if true, 0 if false.",
        "multiline": "Fails if the tweet is not multiple lines. 1 if true, 0 if false. 0 is bad.",
    },
)


## Example of how to create the dataset

if __name__ == "__main__":
    from langsmith import Client

    client = Client()

    topics = [
        "NBA",
        "NFL",
        "Movies",
        "Taylor Swift",
        "Artificial Intelligence",
        "Climate Change",
        "Space Exploration",
        "Cryptocurrency",
        "Healthy Living",
        "Travel Destinations",
        "Technology Trends",
        "Fashion",
        "Food and Cooking",
        "Music Festivals",
        "Entrepreneurship",
        "Fitness",
        "Gaming",
        "Politics",
        "Environmental Conservation",
        "Social Media Trends",
        "Education",
        "Mental Health",
        "Renewable Energy",
        "Virtual Reality",
        "Sustainable Fashion",
        "Robotics",
        "Quantum Computing",
        "Genetic Engineering",
        "Smart Cities",
        "Cybersecurity",
        "Augmented Reality",
        "Electric Vehicles",
        "Blockchain",
        "3D Printing",
        "Nanotechnology",
        "Biotechnology",
        "Internet of Things",
        "Cloud Computing",
        "Big Data",
        "Machine Learning",
        "Artificial General Intelligence",
        "Space Tourism",
        "Autonomous Vehicles",
        "Drones",
        "Wearable Technology",
        "Personalized Medicine",
        "Telemedicine",
        "Remote Work",
        "Digital Nomads",
        "Gig Economy",
        "Circular Economy",
        "Vertical Farming",
        "Lab-grown Meat",
        "Plant-based Diets",
        "Mindfulness",
        "Yoga",
        "Meditation",
        "Biohacking",
        "Nootropics",
        "Intermittent Fasting",
        "HIIT Workouts",
        "Esports",
        "Streaming Services",
        "Podcasting",
        "True Crime",
        "Tiny Houses",
        "Minimalism",
        "Zero Waste Living",
        "Upcycling",
        "Eco-tourism",
        "Voluntourism",
        "Digital Detox",
        "Slow Living",
        "Hygge",
        "Urban Gardening",
        "Permaculture",
        "Regenerative Agriculture",
        "Microplastics",
        "Ocean Conservation",
        "Rewilding",
        "Endangered Species",
        "Biodiversity",
        "Ethical AI",
        "Data Privacy",
        "Net Neutrality",
        "Deepfakes",
        "Fake News",
        "Social Media Activism",
        "Cancel Culture",
        "Meme Culture",
        "NFTs",
        "Decentralized Finance",
        "Universal Basic Income",
        "Gender Equality",
        "LGBTQ+ Rights",
        "Black Lives Matter",
        "Indigenous Rights",
        "Refugee Crisis",
        "Global Poverty",
        "Universal Healthcare",
        "Drug Decriminalization",
        "Prison Reform",
        "Gun Control",
        "Voting Rights",
        "Gerrymandering",
        "Campaign Finance Reform",
        "Term Limits",
        "Ranked Choice Voting",
        "Direct Democracy",
        "Space Debris",
        "Asteroid Mining",
        "Mars Colonization",
        "Extraterrestrial Life",
        "Dark Matter",
        "Black Holes",
        "Quantum Entanglement",
        "Fusion Energy",
        "Antimatter",
        "Cryonics",
        "Life Extension",
        "Transhumanism",
        "Cyborgs",
        "Brain-Computer Interfaces",
        "Memory Implants",
        "Holographic Displays",
    ]

    # Create datasets
    ds = client.create_dataset(dataset_name="tweet-optim")

    # Split topics into train, dev, and test sets
    train_topics = topics[:80]
    dev_topics = topics[80:90]
    test_topics = topics[90:]

    # Create examples for each dataset
    for split_name, dataset_topics in [
        ("train", train_topics),
        ("dev", dev_topics),
        ("test", test_topics),
    ]:
        client.create_examples(
            inputs=[{"topic": topic} for topic in dataset_topics],
            dataset_id=ds.id,
            splits=[split_name] * len(dataset_topics),
        )

    print("Dataset created successfully!")



def under_180_chars(run, example):
    """Evaluate if the tweet is under 180 characters."""
    result = run.outputs.get("tweet", "")
    score = int(len(result) < 180)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "under_180_chars",
        "score": score,
        "comment": comment,
    }


def no_hashtags(run, example):
    """Evaluate if the tweet contains no hashtags."""
    result = run.outputs.get("tweet", "")
    score = int("#" not in result)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "no_hashtags",
        "score": score,
        "comment": comment,
    }


def multiple_lines(run, example):
    """Evaluate if the tweet contains multiple lines."""
    result = run.outputs.get("tweet", "")
    score = int("\n" in result)
    comment = "Pass" if score == 1 else "Fail"
    return {
        "key": "multiline",
        "score": score,
        "comment": comment,
    }


evaluators = [multiple_lines, no_hashtags, under_180_chars]



"""Evaluators to optimize task: example-tweet.

THIS IS A TEMPLATE FOR YOU TO CHANGE!

Evaluators compute scores for prompts run over the configured dataset:
https://smith.langchain.com/o/ebbaf2eb-769b-4505-aca2-d11de10372a4/datasets/db688c10-764b-42ec-acce-4d62419600ed
"""
from langchain_core.messages import AIMessage
from langsmith.schemas import Run, Example

# Modify these evaluators to measure the requested criteria.
# For most prompt optimization tasks, run.outputs["output"] will contain an AIMessage
# (Advanced usage) If you are defining a custom system to optimize, then the outputs will contain the object returned by your system


def example_evaluator(run: Run, example: Example) -> dict:
    """An example evaluator. Larger numbers are better."""
    predicted: AIMessage = run.outputs["output"]

    result = str(predicted.content)
    score = int("#" not in result)
    return {
        "key": "tweet_omits_hashtags",
        "score": score,
        "comment": "Pass: tweet omits hashtags" if score == 1 else "Fail: omit all hashtags from generated tweets",
    }


evaluators = [example_evaluator]



import json
from dataclasses import fields
from typing import Any, Optional, Type

from promptim.trainer import Config
from pydantic import BaseModel, Field, create_model
from pydantic.json_schema import model_json_schema


def get_schema(cls: Type[Any]) -> dict:
    """Create a JSON schema dict from a dataclass or Pydantic model.

    Args:
        cls: A dataclass or Pydantic model type.

    Returns:
        A dict representing the JSON schema of the input class.

    Raises:
        TypeError: If the input is not a dataclass or Pydantic model.
    """
    if isinstance(cls, type) and issubclass(cls, BaseModel):
        return model_json_schema(cls)
    elif hasattr(cls, "__dataclass_fields__"):
        # Convert dataclass to Pydantic model
        fields_dict = {}
        for field in fields(cls):
            field_info = {}
            if field.default is not field.default_factory:
                # Field has a default value or default factory
                field_info["default"] = field.default
            if field.metadata.get("description"):
                field_info["description"] = field.metadata["description"]

            if field_info:
                fields_dict[field.name] = (Optional[field.type], Field(**field_info))
            else:
                # Field is required
                fields_dict[field.name] = (field.type, ...)
        pydantic_model = create_model(cls.__name__, **fields_dict)
        return model_json_schema(pydantic_model)
    else:
        raise TypeError("Input must be a dataclass or Pydantic model")


config_schema = get_schema(Config)
config_schema["$schema"] = "http://json-schema.org/draft-07/schema#"

with open("config-schema.json", "w") as f:
    json.dump(config_schema, f, indent=2)

print("Schema generated and saved to config-schema.json")
